@@ -1,13 +1,12 @@
-/* Time-stamp: <2007-08-07 01:59:57 poser>
+/* Time-stamp: <2008-04-05 10:02:56 poser>
  *
  * Convert text containing various 7-bit ASCII escapes to UTF-7 Unicode.
  *
- * Copyright (C) 2005-2007 William J. Poser (billposer@alum.mit.edu)
+ * Copyright (C) 2005-2008 William J. Poser (billposer@alum.mit.edu)
  *
  * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
+ * it under the terms of version 3 of the GNU General Public License
+ * as published by the Free Software Foundation.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -16,7 +15,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Free Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
  * or go to the web page:  http://www.gnu.org/licenses/gpl.txt.
  */
 
@@ -50,21 +49,21 @@ char compdate[]= "";
 char version[]=PACKAGE_VERSION;
 char pgname[]="ascii2uni";
 
+#define LBUFSIZE 10
+
 #ifndef LOCALEDIR 
 #define LOCALEDIR "/usr/local/share/locale"
 #endif
 
-#define LBUFSIZE 10240
-
 void
 ShowVersion(void){
   fprintf(stderr,"\n%s  %s\n",pgname,version);
 #ifdef HAVE_DATE_TIME
   fprintf(stderr,"%s\n",compdate);
 #endif
-  fprintf(stderr,"Copyright (C) 2004-2007 William J. Poser\n");
+  fprintf(stderr,"Copyright (C) 2004-2008 William J. Poser\n");
   fprintf(stderr,"This program is free software; you can redistribute\n\
-it and/or modify it under the terms of version 2 of\n\
+it and/or modify it under the terms of version 3 of\n\
 the GNU General Public License as published by the\n\
 Free Software Foundation.\n");
   fprintf(stderr,"Report bugs to: billposer@alum.mit.edu.\n");
@@ -79,6 +78,7 @@ containing various representations for non-ASCII characters\nto UTF-8 Unicode.\n
   fprintf(stderr,_("       -a <format specification>.\n"));
   fprintf(stderr,_("       -h Print this usage message.\n"));
   fprintf(stderr,_("       -L List format codes.\n"));
+  fprintf(stderr,_("       -m Accept Microsoft-style HTML entities w/o semi-colon.\n"));
   fprintf(stderr,_("       -p Input consists of pure escapes except for non-null whitespace.\n"));
   fprintf(stderr,_("       -q Quiet - don't chat.\n"));
   fprintf(stderr,_("       -v Print version information.\n"));
@@ -158,14 +158,16 @@ static char *Formats [] = {
 "16#%04X",	/* PSPT */
 "#16r%04X",	/* CLR */
 "16#%04X#",	/* ADA */
-"&#%02X;",	/* HDML */
+"&#02X;",	/* HDML */
 "\\%03o\\%03o\\%03o",	/* BYTEO */
 "\\d%03d\\d%03d\\d%03d",	/* BYTED */
 "\\x%02x\\x%02x\\x%02x",	/* BYTEX */
+"_x%4lX_",	/* OOXML */
 "&%[abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789];", /* CHENT */
 "=%2lX", 		/* UTF-8  - Ifmt */
 "%%%2lX", 		/* UTF-8  - Jfmt */
-"\\%3lo" 		/* UTF-8  - Kfmt */
+"\\%3lo", 		/* UTF-8  - Kfmt */
+"\\x%2lX" 		/* UTF-8  - APACHE */
 };
 
 
@@ -182,13 +184,15 @@ int main (int ac, char *av[])
   UTF32 num;
   int oc;			/* Command line option flag */
   int Converted;
-  long TokenNumber;
-  long ReplacedNumber;
-  short BMPSplitP = 0; 
+  long TokenNumber = 0L;
+  long ReplacedNumber = 0L;
+  short BMPSplitP = 0;
   int VerboseP = 1;
   int UTF8ValueP = 0;		/* Are incoming values UTF-8? */
   short AllHTMLP = 0; 		/* Translate all three kinds of HTML escape */
   int PureP = 0;
+  int StrictP = 1;		/* Do not convert Microsoft-style numeric character references */
+  long MicrosoftStyle = 0L;	/* Number of Microsoft-style ncrs detected */
   int Word_Length;
   int NConsumed;
   int LineNo;
@@ -196,25 +200,26 @@ int main (int ac, char *av[])
   char *iptr;
   int eof;
   char SplitStr[3];
-  char enam[LBUFSIZE];
-  long MicrosoftStyle = 0L;
+  char enam[16];
   char tmpstr [16];
   unsigned char b1;		/* Used for byte-wise encoding */
   unsigned char b2;
   unsigned char b3;
-  int FType;
+  int FType = STDX;
   int UTF8Type;			/* Not used - for compatibility with uni2ascii */
   short UseEntitiesP;	/* Not used - for compatibility with uni2ascii */
   int last;
+  short EndedInNLP = 0;
 
-  char lbuf [LBUFSIZE+1];
+  char lbuf[LBUFSIZE];
 
   extern int optind;
   extern int opterr;
   extern int optopt;
+  extern unsigned long GetWordLineNo;
+
   extern void putu8 (unsigned long);
   extern char * Get_Word(FILE *, int *, int *);
-
   extern int CountSlots(char *);
   extern void ListFormatArguments(short);
   extern void SetFormat(char *, int *, short *,int *, short *, short *);
@@ -231,7 +236,7 @@ int main (int ac, char *av[])
 
   /* Handle command line arguments */
 
-   while( (oc = getopt(ac,av,":Z:a:hLpqv")) != EOF){
+   while( (oc = getopt(ac,av,":Z:a:hLmpqv")) != EOF){
      switch(oc){
      case 'a':
        SetFormat(optarg,&FType,&UseEntitiesP, &UTF8Type, &BMPSplitP,&AllHTMLP);
@@ -240,9 +245,10 @@ int main (int ac, char *av[])
 	 exit(BADOPTIONARG);
        }
        fmt = Formats[FType];
-       if((FType == IFMT) || (FType == JFMT) || (FType == KFMT)) UTF8ValueP =1;
+       if((FType == IFMT) || (FType == JFMT) || (FType == KFMT) || (FType == APACHE)) UTF8ValueP =1;
        if(FType == JFMT) {cbuf[0] = '0'; cbuf[1] = 'x';}
        if(FType == KFMT) {cbuf[0] = '\\';}
+       if(FType == APACHE) {cbuf[0] = '0';}
        break;
     case 'L':
       ListFormatArguments(0);
@@ -250,10 +256,13 @@ int main (int ac, char *av[])
      case 'Z':
        fmt = optarg;
        if(CountSlots(fmt) > 1) {
-	 fprintf(stderr,"You may not supply a format with more than one empty slot.\n");
+	 fprintf(stderr,_("You may not supply a format with more than one empty slot.\n"));
 	 exit(BADOPTIONARG);
        }
        break;
+     case 'm':
+       StrictP = 0;
+       break;
      case 'p':
        PureP = 1;
        break;
@@ -295,8 +304,6 @@ int main (int ac, char *av[])
    }
 
    sprintf(afmt,"%s%%n",fmt);	/* Add %n for NConsumed */
-   ReplacedNumber = 0L;
-   TokenNumber = 0L;
 
    /*
     * This is the case in which the input consists entirely of escapes
@@ -304,14 +311,15 @@ int main (int ac, char *av[])
     */
 
    if(PureP) {
+     GetWordLineNo = 1;
      while(1){
        str = Get_Word(stdin,&Word_Length,&eof);
        if(eof) break; 
        if(Word_Length == 0) continue;
        TokenNumber++;
        if(str == NULL){
-	 fprintf(stderr,_("%1$s: failed to allocate storage for input token %2$ld.\n"),
-		 pgname,TokenNumber);
+	 fprintf(stderr,_("%1$s: failed to allocate storage for input token %2$ld at line %3$ul.\n"),
+		 pgname,TokenNumber,GetWordLineNo);
 	 exit(OUTOFMEMORY);
        }
        if(FType == CHENT) {
@@ -319,8 +327,8 @@ int main (int ac, char *av[])
 	 num = LookupCodeForEntity(enam);
 	 if(!num) {
 	   num = UNI_REPLACEMENT_CHAR;
-	   fprintf(stderr,"ascii2uni: unknown HTML character entity \"&%s;\"\n",
-		   enam);
+	   fprintf(stderr,"ascii2uni: unknown HTML character entity \"&%s;\" at line %lu\n",
+		   enam,GetWordLineNo);
 	   ReplacedNumber++;
 	   Converted = (-1);
 	 }
@@ -349,23 +357,34 @@ int main (int ac, char *av[])
        }
 
        if(Converted < 1) {
-	 fprintf(stderr,_("Ill-formed input %1$s at token %2$lu\n"),str,TokenNumber);
+	 fprintf(stderr,_("Ill-formed input %1$s at token %2$lu at line %3$lu\n"),
+		 str,TokenNumber,GetWordLineNo);
 	 exit(BADRECORD); 
        }
        else if(Converted > 3) {
-	 fprintf(stderr,_("The character encoded as %1$s at token %2$lu is outside the Unicode range.\n\tEmitting Unicode replacement character.\n"),
-		 str,TokenNumber);
+	 fprintf(stderr,_("The character encoded as %1$s at token %2$lu at line %3$lu is outside the Unicode range.\n\tEmitting Unicode replacement character.\n"),
+		 str,TokenNumber,GetWordLineNo);
 	 putu8(UNI_REPLACEMENT_CHAR);
        } 
        else {
-	 if (UTF8ValueP) putchar(num);
-	 else putu8(num);
 	 if( (FType == HTMLD) || (FType == HTMLX) || (FType == CHENT) || (FType == HDML)) {
 	   if(*(str+NConsumed-1) != ';') {
 	     MicrosoftStyle++;
-	     fprintf(stderr,_("The HTML/HDML entity %1$s at token %2$lu lacks the requisite final semicolon.\n"),str,TokenNumber);
+	     fprintf(stderr,_("The HTML/HDML entity %1$s at token %2$lu at line %3$lu lacks the requisite final semicolon.\n"),
+		     str,TokenNumber,GetWordLineNo);
+	     if(StrictP) {
+	       fputs(str,stdout);
+	       TokenNumber--;
+	     } else {
+	       if (UTF8ValueP) putchar(num);
+	       else putu8(num);
+	     }
+	     free((void *)str);
+	     continue;
 	   }
 	 }
+	 if (UTF8ValueP) putchar(num);
+	 else putu8(num);
        }
        free((void *)str);
      }
@@ -373,15 +392,16 @@ int main (int ac, char *av[])
    } /* End of PureP */
 
    /* This is the case in which the Unicode escapes are embedded in ASCII text */
-
    LineNo = 0;
    while(fgets(lbuf,LBUFSIZE,stdin) != NULL) {
-     LineNo++;
+     EndedInNLP = 0;
      last = strlen(lbuf) - 1;
      if(lbuf[last] == '\n') {
+       EndedInNLP = 1;
+       LineNo++;
        lbuf[last] = '\0';
        last--;
-     }
+     } 
      if(last < 0) {
        putchar('\n');
        continue;
@@ -455,21 +475,45 @@ int main (int ac, char *av[])
        else if (FType == CHENT) {
 	 if (AllHTMLP){
 	   if(sscanf(iptr,aHfmt,&num,&NConsumed)) {
-	     putu8(num);
-	     iptr+=NConsumed;
-	     if(*(iptr-1) != ';') {
+	     if(*(iptr+NConsumed-1) != ';') {
 	       MicrosoftStyle++;
-	       fprintf(stderr,_("The HTML/HDML entity %1$s at token %2$lu lacks the requisite final semicolon.\n"),ExtractSubstring(tmpstr,iptr-NConsumed,iptr-1),TokenNumber);
+	       fprintf(stderr,
+		       _("The HTML/HDML entity %1$s at token %2$lu of line %3$lu lacks the requisite final semicolon.\n"),
+		       ExtractSubstring(tmpstr,iptr,iptr+NConsumed-3),TokenNumber,LineNo);
+	       if(StrictP) {
+		 putchar(*iptr++);
+		 continue;
+	       }
+	       else {
+		 putu8(num);
+		 iptr+=NConsumed;
+	       }
+	     }
+	     else {
+	       putu8(num);
+	       iptr+=NConsumed;
 	     }
 	     TokenNumber++;
 	     continue;
 	   }
 	   else if(sscanf(iptr,aDfmt,&num,&NConsumed)) {
-	     putu8(num);
-	     iptr+=NConsumed;
-	     if(*(iptr-1) != ';') {
+	     if(*(iptr+NConsumed-1) != ';') {
 	       MicrosoftStyle++;
-	       fprintf(stderr,_("The HTML/HDML entity %1$s at token %2$lu lacks the requisite final semicolon.\n"),ExtractSubstring(tmpstr,iptr-NConsumed,iptr-1),TokenNumber);
+	       fprintf(stderr,
+		       _("The HTML/HDML entity %1$s at token %2$lu of line %3$lu lacks the requisite final semicolon.\n"),
+		       ExtractSubstring(tmpstr,iptr,iptr+NConsumed-3),TokenNumber,LineNo);
+	       if (StrictP) {
+		 putchar(*iptr++);
+		 continue;
+	       }
+	       else {
+		 putu8(num);
+		 iptr+=NConsumed;
+	       }
+	     }
+	     else {
+	       putu8(num);
+	       iptr+=NConsumed;
 	     }
 	     TokenNumber++;
 	     continue;
@@ -477,11 +521,21 @@ int main (int ac, char *av[])
 	 }
 	 if(sscanf(iptr,afmt,&enam,&NConsumed)) {
 	   if( (num = LookupCodeForEntity(enam))) {
-	     putu8(num);
-	     iptr+=NConsumed;
-	     if(*(iptr-1) != ';') {
+	     if(*(iptr+NConsumed-1) != ';') {
 	       MicrosoftStyle++;
-	       fprintf(stderr,_("The HTML/HDML entity %1$s at token %2$lu lacks the requisite final semicolon.\n"),ExtractSubstring(tmpstr,iptr-NConsumed,iptr-1),TokenNumber);
+	       fprintf(stderr,_("The HTML/HDML entity %1$s at token %2$lu of line %3$lu lacks the requisite final semicolon.\n"),ExtractSubstring(tmpstr,iptr,iptr+NConsumed-3),TokenNumber,LineNo);
+	       if(StrictP) {
+		 putchar(*iptr++);
+		 continue;
+	       }
+	       else {
+		 putu8(num);
+		 iptr+=NConsumed;
+	       }
+	     }
+	     else {
+	       putu8(num);
+	       iptr+=NConsumed;
 	     }
 	     TokenNumber++;
 	   }
@@ -497,6 +551,7 @@ int main (int ac, char *av[])
        } /* End of Qfmt case */
        else if( (BYTEO == FType) || (BYTED == FType) || (BYTEH == FType)) {
 	 Converted=sscanf(iptr,afmt,&b1,&b2,&b3,&NConsumed);
+	 fprintf(stderr,"Converted = %d\n",Converted);fflush(stderr);
 	 switch(Converted)
 	   {
 	   case 3:
@@ -515,40 +570,52 @@ int main (int ac, char *av[])
 	     putchar(*iptr++);
 	     break;
 	   default:
-	     fprintf(stderr,_("The character encoded as %1$s at token %2$lu is outside the Unicode range.\n\tEmitting Unicode replacement character.\n"),
-		     str,TokenNumber);
+	     fprintf(stderr,
+		     _("The character encoded as %1$s at token %2$lu of line %3$lu is outside the Unicode range.\n\tEmitting Unicode replacement character.\n"),
+		     str,TokenNumber,LineNo);
 	     putu8(UNI_REPLACEMENT_CHAR);
-	   }
+	   } /* end switch */
 	   TokenNumber++;
        }
        else if (HDML == FType) {
-	 /* Check for HDML character entities */
+	 /* HDML character references */
+	 /* Need to fill this in */
        }
        else {			/* Default - not BMPSplitP, Q, or byte format */
-	 //	 NConsumed=0;
 	 if(sscanf(iptr,afmt,&num,&NConsumed)) {
-	   if (UTF8ValueP) putchar(num);
-	   else putu8(num);
-	   iptr+=NConsumed;
 	   if(FType== HTMLX) {
-	     if(*(iptr-1) != ';') {
+	     if(*(iptr-1+NConsumed) != ';') {
 	       MicrosoftStyle++;
-	       /*	       fprintf(stderr,_("The HTML entity %1$s at token %2$lu lacks the requisite final semicolon.\n"),ExtractSubstring(tmpstr,iptr-NConsumed,iptr-1),TokenNumber); */
-	       fprintf(stderr,"The HTML entity %s at token %lu lacks the requisite final semicolon.\n",ExtractSubstring(tmpstr,iptr-NConsumed,iptr-1),TokenNumber);
+	       fprintf(stderr,
+		       "The HTML entity %1$s at token %2$lu of line %3$d lacks the requisite final semicolon.\n",
+		       ExtractSubstring(tmpstr,iptr,iptr+NConsumed-3),TokenNumber,LineNo);
+	       if(StrictP) {
+		 putchar(*iptr++);
+		 continue;
+	       }
 	     }
 	   }
 	   else if(FType == HTMLD) {
-	     if(*(iptr-1) != ';') {
+	     if(*(iptr-1+NConsumed) != ';') {
 	       MicrosoftStyle++;
-	       fprintf(stderr,_("The HTML entity %1$s at token %2$lu lacks the requisite final semicolon.\n"),ExtractSubstring(tmpstr,iptr-NConsumed,iptr-1),TokenNumber);
+	       fprintf(stderr,
+		       _("The HTML entity %1$s at token %2$lu of line %3$d lacks the requisite final semicolon.\n"),
+		       ExtractSubstring(tmpstr,iptr,iptr+NConsumed-3),TokenNumber,LineNo);
+	       if(StrictP) {
+		 putchar(*iptr++);
+		 continue;
+	       }
 	     }
 	   }
+	   if (UTF8ValueP) putchar(num);
+	   else putu8(num);
+	   iptr+=NConsumed;
 	   TokenNumber++;
-	 }
+	 } /* End of if(sscanf */
 	 else putchar(*iptr++);
        }
      } /* Loop over current line */
-     putchar('\n');
+     if(EndedInNLP) putchar('\n');
    } /* Loop over input lines */
 
 done:
@@ -561,10 +628,15 @@ done:
        else fprintf(stderr,_("%ld tokens replaced with Unicode Replacement Character\n"),ReplacedNumber);
      }
      if(MicrosoftStyle) {
+       if(StrictP) {
+	 fprintf(stderr,
+	       _("%ld Microsoft-style (lacking final semi-colon) not converted\n"),MicrosoftStyle);
+       }
+       else {
        fprintf(stderr,
-	       _("%ld Microsoft-style (lacking final semi-colon)\n"),MicrosoftStyle);
+	       _("%ld Microsoft-style (lacking final semi-colon) among those converted\n"),MicrosoftStyle);
+       }
      }
    }
    exit(SUCCESS);
 }
-
