@@ -1,14 +1,13 @@
-/* Time-stamp: <2007-08-07 00:54:51 poser>
+/* Time-stamp: <2008-04-26 11:26:17 poser>
  *
  * Converts UTF-8 Unicode to pure 7-bit ASCII using any of a number
  * of different representations. 
  * 
- * Copyright (C) 2004-2007 William J. Poser (billposer@alum.mit.edu)
+ * Copyright (C) 2004-2008 William J. Poser (billposer@alum.mit.edu)
  *
  * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
+ * it under the terms of version 3 of the GNU General Public License as published by
+ * the Free Software Foundation.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -17,7 +16,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Free Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
  * or go to the web page:  http://www.gnu.org/licenses/gpl.txt.
  */
 
@@ -67,9 +66,9 @@ ShowVersion(void){
 #ifdef HAVE_DATE_TIME
   fprintf(stderr,"%s\n",compdate);
 #endif
-  fprintf(stderr,"Copyright (C) 2004-2007 William J. Poser\n");
+  fprintf(stderr,"Copyright (C) 2004-2008 William J. Poser\n");
   fprintf(stderr,"This program is free software; you can redistribute\n\
-it and/or modify it under the terms of version 2 of\n\
+it and/or modify it under the terms of version 3 of\n\
 the GNU General Public License as published by the\n\
 Free Software Foundation.\n");
   fprintf(stderr,"Report bugs to: billposer@alum.mit.edu.\n");
@@ -95,13 +94,16 @@ and for conversion even of newline and space characters.\n"));
   fprintf(stderr,_("       -e Convert to approximate ASCII equivalents.\n"));
   fprintf(stderr,_("       -f Convert formal equivalents to ASCII.\n"));
   fprintf(stderr,_("       -x Expand certain single characters to equivalent sequences.\n"));
+  fprintf(stderr,_("       -y Convert certain characters to approximate single-characters.\n"));
   fprintf(stderr,_("       -d Strip diacritics.\n"));
   fprintf(stderr,_("       -S <Unicode:ASCII>  define a custom substitution\n"));
   fprintf(stderr,_("       -n Convert newlines.\n"));
   fprintf(stderr,_("       -s Convert space characters.\n"));
+  fprintf(stderr,_("       -P Pass through Unicode if not transformed - do not ascify\n"));
   fprintf(stderr,_("       -p Convert codepoints below 0x80 too.\n"));
   fprintf(stderr,_("       -l Use lower-case a-f when generating hex.\n"));
   fprintf(stderr,_("       -w Add a space after each converted item.\n"));
+  fprintf(stderr,_("       -B Best ASCII. Transform to ASCII if possible. Combines cdefx.\n"));
   fprintf(stderr,
 	  _("       -Z <format> Use the specified format\n"));
   fprintf(stderr,_("Report bugs to: billposer@alum.mit.edu\n"));
@@ -135,6 +137,46 @@ void ListExpansions(void){
   fprintf(stderr,_("          U+21D0 LEFTWARDS DOUBLE ARROW      -> <=\n"));
   fprintf(stderr,_("          U+21D2 RIGHTWARDS DOUBLE ARROW     -> =>\n"));
   fprintf(stderr,_("          U+22EF MIDLINE HORIZONTAL ELLIPSIS -> ...\n"));
+  fprintf(stderr,_("          U+FB00 LATIN SMALL LIGATURE FF -> ff\n"));
+  fprintf(stderr,_("          U+FB01 LATIN SMALL LIGATURE FI -> fi\n"));
+  fprintf(stderr,_("          U+FB02 LATIN SMALL LIGATURE FL -> fl\n"));
+  fprintf(stderr,_("          U+FB03 LATIN SMALL LIGATURE FFI -> ffi\n"));
+  fprintf(stderr,_("          U+FB04 LATIN SMALL LIGATURE FFL -> ffl\n"));
+  fprintf(stderr,_("          U+FB06 LATIN SMALL LIGATURE ST -> st\n"));
+}
+
+void ListSingleApproximations(void){
+  fprintf(stderr,_("The approximations produced by the -y flag are:\n"));
+  fprintf(stderr,_("          U+00A2 CENT SIGN                   -> C\n"));
+  fprintf(stderr,_("          U+00A3 POUND SIGN                  -> #\n"));
+  fprintf(stderr,_("          U+00A5 YEN SIGN                    -> Y\n"));
+  fprintf(stderr,_("          U+00A9 COPYRIGHT SYMBOL            -> C\n"));
+  fprintf(stderr,_("          U+00AE REGISTERED SYMBOL           -> R\n"));
+  fprintf(stderr,_("          U+00AF MACRON                      -> -\n"));
+  fprintf(stderr,_("          U+00BC ONE QUARTER                 -> -\n"));
+  fprintf(stderr,_("          U+00BD ONE HALF                    -> -\n"));
+  fprintf(stderr,_("          U+00BE THREE QUARTERS              -> -\n"));
+  fprintf(stderr,_("          U+00C6 CAPITAL LETTER ASH          -> A\n"));
+  fprintf(stderr,_("          U+00DF SMALL LETTER SHARP S        -> s\n"));
+  fprintf(stderr,_("          U+00E6 SMALL LETTER ASH            -> a\n"));
+  fprintf(stderr,_("          U+0132 LIGATURE IJ                 -> I\n"));
+  fprintf(stderr,_("          U+0133 LIGATURE ij                 -> i\n"));
+  fprintf(stderr,_("          U+0152 LIGATURE OE                 -> O\n"));
+  fprintf(stderr,_("          U+0153 LIGATURE oe                 -> o\n"));
+  fprintf(stderr,_("          U+01F1 CAPITAL LETTER DZ           -> D\n"));
+  fprintf(stderr,_("          U+01F2 MIXED LETTER Dz             -> D\n"));
+  fprintf(stderr,_("          U+01F3 SMALL LETTER dz             -> d\n"));
+  fprintf(stderr,_("          U+02A6 SMALL LETTER TS DIGRAPH     -> t\n"));
+  fprintf(stderr,_("          U+2026 HORIZONTAL ELLIPSIS         -> .\n"));
+  fprintf(stderr,_("          U+20AC EURO SIGN                   -> E\n"));
+  fprintf(stderr,_("          U+2190 LEFTWARDS ARROW             -> <\n"));
+  fprintf(stderr,_("          U+2192 RIGHTWARDS ARROW            -> >\n"));
+  fprintf(stderr,_("          U+21D0 LEFTWARDS DOUBLE ARROW      -> <\n"));
+  fprintf(stderr,_("          U+21D2 RIGHTWARDS DOUBLE ARROW     -> >\n"));
+  fprintf(stderr,_("          U+2216 SET MINUS                   -> \\\n"));
+  fprintf(stderr,_("          U+2217 ASTERISK OPERATOR           -> *\n"));
+  fprintf(stderr,_("          U+2223 DIVIDES                     -> |\n"));
+  fprintf(stderr,_("          U+22EF MIDLINE HORIZONTAL ELLIPSIS -> .\n"));
 }
 
 /* Swap the byte order of a 4 byte integer */
@@ -206,14 +248,16 @@ static char *Formats [] = {
   "#16r%04X",
   "16#%04x#",	/* ADA */
   "16#%04X#",
-  "&#%02x;",  /* HDML */
+  "&#%02x;",   /* HDML */ 
   "&#%02X;",
   "\\%03hho\\%03hho\\%03hho",	/* BYTEO */
   "\\%03hho\\%03hho\\%03hho",
   "\\d%03hhd\\d%03hhd\\d%03hhd",	/* BYTED */
   "\\d%03hhd\\d%03hhd\\d%03hhd",
   "\\x%02hhx\\x%02hhx\\x%02hhx",/* BYTEH */
-  "\\x%02hhX\\x%02hhX\\x%02hhX"
+  "\\x%02hhX\\x%02hhX\\x%02hhX",
+  "_x%04x_", /* OOXML */
+  "_x%04X_" 
   };
 
 /*  Returns 1 if it handled the character, 0 if not. */
@@ -415,6 +459,9 @@ int AscifyEnclosed(UTF32 c) {
 int AscifyEquiv(UTF32 c) {
   switch (c)
     {
+    case 0x00AF:
+      putchar('-');		/* macron */
+      break;
     case 0xFE60:
       putchar(0x26);		/* ampersand */
       break;
@@ -482,6 +529,15 @@ int AscifyEquiv(UTF32 c) {
     case 0x3000:		/* ideographic space */
       putchar(0x20);		/* Space */
       break;
+    case 0x2215:		/* division slash */
+      putchar(0x2F);
+      break;
+    case 0x2216:		/* set minus */
+      putchar(0x5C);
+      break;
+    case 0x2223:		/* divides */
+      putchar('|');
+      break;
     default:
       return 0;
     }
@@ -2342,6 +2398,87 @@ int AscifyStyle(UTF32 c) {
 } /* End of AscifyStyle */
 
 
+int MakeAscii(UTF32 c) {
+  switch(c)
+    {
+    case 0x00A2:	/* Cent sign */
+      putchar('c');
+      break;
+    case 0x00A3:	/* Pound sign */
+      putchar('#');
+      break;
+    case 0x00A5:	/* Yen sign */
+      putchar('Y');
+      break;
+    case 0x00A9:	/* Copyright symbol */
+      putchar('C');
+      break;
+    case 0x00AE:	/* Registered symbol */
+      putchar('R');
+      break;
+    case 0x00BC:	/* One quarter */
+    case 0x00BD:	/* One half */
+    case 0x00BE:	/* Three quarters */
+      putchar('-');
+      break;
+    case 0x00C6:	/* Ash */
+      putchar('A');
+      break;
+    case 0x00E6:	/* ash */
+      putchar('a');
+      break;
+    case 0x00DF:	/* Eszet */
+      putchar('s');
+      break;
+    case 0x0132:	/* Ligature IJ */
+      putchar('I');
+      break;
+    case 0x0133:	/* Ligature ij */
+      putchar('i');
+      break;
+    case 0x0152:	/* Ligature OE */
+      putchar('O');
+      break;
+    case 0x0153:	/* Ligature oe */
+      putchar('o');
+      break;
+    case 0x01F1:	/* Capital DZ */
+      putchar('D');
+      break;
+    case 0x01F2:	/* Mixed Dz */
+      putchar('D');
+      break;
+    case 0x01F3:	/* Small DZ */
+      putchar('d');
+      break;
+    case 0x02A6:	/* Small ts digraph */
+      putchar('t');
+      break;
+    case 0x20AC:	/* Euro sign */
+      putchar('E');
+      break;
+    case 0x2026:	/* Horizontal ellipsis */
+    case 0x22EF:	/* Midline horizontal ellipsis */
+      putchar('.');
+      break;
+    case 0x2190:	/* Leftwards arrow */
+      putchar('<');
+      break;
+    case 0x2192:	/* Rightwards arrow */
+      putchar('>');
+      break;
+    case 0x21D0:	/* Leftwards double arrow */
+      putchar('<');
+      break;
+    case 0x21D2:	/* Rightwards double arrow */
+      putchar('>');
+      break;
+    default:
+      return 0;
+  }
+  return 1;
+}
+
 int ExpandToAscii(UTF32 c) {
   switch(c)
     {
@@ -2460,6 +2597,32 @@ int ExpandToAscii(UTF32 c) {
       putchar('=');
       putchar('>');
       break;
+    case 0xFB00:
+      putchar('f');
+      putchar('f');
+      break;
+    case 0xFB01:
+      putchar('f');
+      putchar('i');
+      break;
+    case 0xFB02:
+      putchar('f');
+      putchar('l');
+      break;
+    case 0xFB03:
+      putchar('f');
+      putchar('f');
+      putchar('i');
+      break;
+    case 0xFB04:
+      putchar('f');
+      putchar('f');
+      putchar('l');
+      break;
+    case 0xFB06:
+      putchar('s');
+      putchar('t');
+      break;
     default:
       return 0;
   }
@@ -2581,12 +2744,14 @@ int main (int ac, char *av[])
   short VerboseP = 1;
   short HexUpperP = 1;	/* Use X or x in  hex ?  */
   short AddWhitespaceP = 0;
+  short PassThroughP = 0;	/* If untransformed, pass through unicode rather than asciifying */
 
   short AscifyStyleP = 0;
   short AscifyDiacriticsP = 0;
   short AscifyEnclosedP = 0;
   short AscifyEquivP = 0;
   short ExpandP = 0;
+  short MakeP = 0;
   short LittleEndianP = 0;
 
   char *AboveBMPfmt;
@@ -2621,7 +2786,7 @@ int main (int ac, char *av[])
 
   /* Handle command line arguments */
 
-  while( (oc = getopt(ac,av,":a:cdeEfhlLnpqsS:vwxZ:")) != EOF){
+  while( (oc = getopt(ac,av,":Aa:BcdeEfhlLnPpqsS:vwyxZ:")) != EOF){
     switch(oc){
     case 'a':
       SetFormat(optarg,&FType,&UseEntitiesP, &UTF8Type, &BMPSplitP,&dummy);
@@ -2630,6 +2795,16 @@ int main (int ac, char *av[])
 	 exit(BADOPTIONARG);
        }
       break;
+    case 'A':
+      ListSingleApproximations();
+      exit(INFO);
+    case 'B':
+      AscifyEnclosedP = 1;
+      AscifyDiacriticsP = 1;
+      AscifyEquivP = 1;
+      AscifyStyleP = 1;
+      ExpandP = 1;
+      break;
     case 'c':
       AscifyEnclosedP = 1;
       break;
@@ -2654,6 +2829,10 @@ int main (int ac, char *av[])
     case 'n':
       PreserveNewlinesP = 0;
       break;
+    case 'P':
+      PassThroughP = 1;
+      UTF8Type = 0;
+      break;
     case 'p':
       PureP =1;
       break;
@@ -2666,6 +2845,9 @@ int main (int ac, char *av[])
     case 'x':
       ExpandP = 1;
       break;
+    case 'y':
+      MakeP = 1;
+      break;
     case 'w':
       AddWhitespaceP = 1;
       break;
@@ -2739,7 +2921,7 @@ int main (int ac, char *av[])
   }
 
   if(UTF8Type) {
-    if(AscifyStyleP|AscifyDiacriticsP|AscifyEquivP|AscifyEnclosedP|ExpandP) {
+    if(AscifyStyleP|AscifyDiacriticsP|AscifyEquivP|AscifyEnclosedP|ExpandP|MakeP) {
       while ( (c = UTF8in(infd,&UCBytes,&ustr)) <= UNI_MAX_UTF32) {
 	ByteCnt+=UCBytes;
 	CharCnt++;
@@ -2772,6 +2954,12 @@ int main (int ac, char *av[])
 	    continue;
 	  }
 	}
+	if(MakeP) {
+	  if(MakeAscii(c)) {
+	    SubstitutedCnt++;
+	    continue;
+	  }
+	}
 	if(AscifyEquivP) {
 	  if(AscifyEquiv(c)) {
 	    SubstitutedCnt++;
@@ -2800,6 +2988,9 @@ int main (int ac, char *av[])
 	    case 3:
 	      printf("\\%03o",ch);
 	      break;
+	    case 4:
+	      printf("\\x%02X",ch);
+	      break;
 	    }
 	  }
 	  if(AddWhitespaceP) putchar(' ');
@@ -2822,6 +3013,9 @@ int main (int ac, char *av[])
 	  case 3:
 	    printf("\\%03o",ch);
 	    break;
+	  case 4:
+	    printf("\\x%02X",ch);
+	    break;
 	  }
 	}
 	if(AddWhitespaceP) putchar(' ');
@@ -2862,6 +3056,12 @@ int main (int ac, char *av[])
 	continue;
       }
     }
+	if(MakeP) {
+	  if(MakeAscii(c)) {
+	    SubstitutedCnt++;
+	    continue;
+	  }
+	}
     if(AscifyEquivP) {
       if(AscifyEquiv(c)) {
 	SubstitutedCnt++;
@@ -2910,7 +3110,8 @@ int main (int ac, char *av[])
       }
       break;
     default:
-      if(!PureP && (c <= 0x7F)) putchar((int)c);
+      if(PassThroughP) putu8(c);
+      else if(!PureP && (c <= 0x7F)) putchar((int)c);
       else {
 	if(BMPSplitP) {
 	  if(c > 0xFFFF) printf(AboveBMPfmt,c);
